#!/usr/bin/env ruby

=begin
 sqid   -- SQL Injection Digger
           http://sqid.rubyforge.org

 Copyright (C) Metaeye Security Group <contact@metaeye.org>

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License version 2 as
 published by the Free Software Foundation; version 2.

 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 A copy of the GNU GPL is available as /usr/doc/copyright/GPL on Debian
 systems, or on the World Wide Web at http://www.gnu.org/copyleft/gpl.html
 You can also obtain it by writing to the Free Software Foundation, Inc.,
 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
=end

# program name and version
$program = ["sqid" , "0.1"]

# defualt search query
$search_query=nil

# defualt where to start
$start=0

# defualt number of searches to perform
$no_of_searches=50

# The file containing signatures.
$sqid_db_file = 'sqid.db'


puts "#{$program.join(" v")} - SQL Injection digger (Google search mode).\n"
puts "Copyright (C) Metaeye Security Group - http://sqid.rubyforge.org\n\n"


# Look for google key file or GOOGLE_LICENSE_KEY environment variable
# exit if not found
$google_key=nil

begin
    $google_key = File.open("#{ENV['HOME']}/.google_key") {|kf| kf.readline.chomp} 
rescue
end

$google_key = ENV['GOOGLE_LICENSE_KEY'] if ! $google_key

if  ! $google_key
    puts "[-] Google key for search operations not found."
    puts "    Either create a file in home directory named .google_key,"
    puts "    or set the environment variable GOOGLE_LICENSE_KEY."
    exit(1)
end



def print_usage()
usage= <<USAGE_STRING
usage: #{$program[0]} <search query> [ <start> [ <results>]]

    <search query> : the query to perform google search for.
    <start> : zero-based index of the first desired result,is
              zero if not specified.
    <results> : number of results desired, is 50 if not specfied.
USAGE_STRING
    puts usage
end

# Try to load the require modules, bail out if not found
begin
    require 'http-access2'
    require 'google'
rescue LoadError
    puts "[!] Failed to load required modules, " + $!.message
    exit(1)
end


class Sqid
    
    def initialize(db_file,google_key)
	
	# This hash holds the errors and their regex
	@SQLErrors=Hash.new()
	
	# This hash holds the urls checked and thier params
	@Checkedurls=Hash.new()
	
	# The google key for searching
	@Googlekey=google_key
	
	# This holds the results of a query
	@SearchResults=Array.new()
	
	begin
	    File.open(db_file,"r") do |file|
    		while line=file.gets
		    line=line.strip()
		    if line =~ /^#/ || line == ""
			next
		    end
		    a=line.split("|",2)
		    @SQLErrors.store( Regexp.new( a[0].strip() ) , a[1].strip() )
		end
	    end
	rescue Errno::ENOENT
	    puts "[!] Failed to load signatures, " + $!.message
	    exit(1)	
	rescue
	    puts "[!] Failed to load signatures, corrupted signatures file - #{db_file}."
	    exit(1)
	end
    end

    # Need to make this clean and better
    # This gives us case urls for test for ex. if url is http://www.foo.com/foo?a=1&b=1 it gives
    #   http://www.foo.com/foo?a='&b=1
    #   http://www.foo.com/foo?a=1&b='
    # Also checks for duplicates
    
    def get_test_urls(url)
	t=url.split("?",2)
	
	if ! @Checkedurls[t[0]]
	    @Checkedurls.store(t[0],"")
	end
	
	if t[1]
	    a=t[1].split("&")
	    if a
		j=0
		while j < a.length()
		    s=a[j]
		    parm=a[j].split("=")[0]
		    if @Checkedurls[t[0]].split("&").index(parm)
			j+=1
			next
		    end
		    a[j]=parm+"='" if a[j].index("=")
		    @Checkedurls[t[0]]+=(parm+"&")
		    yield t[0]+"?"+a.join("&")
		    a[j]=s
		    j+=1
		end
	    end
	end
    end

    # Check and report an error in a url
    def check_url(url)
	proxy = ENV['HTTP_PROXY']
	clnt = HTTPAccess2::Client.new(proxy)

    	get_test_urls(url) { |test_url|
	    match=false
	    response=clnt.get(test_url)
	    resp=String.new(response.content).split("\n").join("") # Ruby regex won't match if there is a newline.
	    @SQLErrors.each do |key,value|
		if resp =~ /#{key}/
		    puts String(response.status) + " #{value} => " + test_url
		    match=true
		end
	    end
	    # Just a lucky test
	    if match == false && response.status == 500
		    puts String(response.status) + " No Match => " + test_url
	    end
	}
    end
	
    def check_urls(urls)
	urls.each { |url| check_url(url) }
    end
	
    def upto_in_10s(num)
	r=Array.new()
	while num > 0
	    r.push(10)
	    num=num-10
	end
	r.push(num)
	return r
    end
	
    def do_search(search_query,start,no_of_searches)
	google = Google::Search.new(@Googlekey)
	upto_in_10s(no_of_searches).each do |maxResults|
	    res = google.search(search_query,start,maxResults)
	    res.resultElements.each { |result| @SearchResults.push(result.url) }
	    start+=maxResults
	end
	return @SearchResults
    end

    def search_and_find(search_query,start,no_of_searches)
	check_urls(do_search(search_query,start,no_of_searches))
    end
end


# Parse command line arguments
begin
    case ARGV.length
	when 1
	    $search_query=ARGV[0]
	when 2
	    $search_query=ARGV[0]
	    $start=Integer(ARGV[1])
	when 3
	    $search_query=ARGV[0]
	    $start=Integer(ARGV[1])
	    $no_of_searches=Integer(ARGV[2])
   	else
	    print_usage()
	    exit(1)
    end
rescue ArgumentError
    puts "[-] Invalid argument supplied, " + $!.message
end
# end parsing command line arguments

printf "[+] Performing #{$no_of_searches} search(s) for '#{$search_query}' starting from #{$start}.\n\n"

begin
    # Clean and simple
    s=Sqid.new($sqid_db_file,$google_key)
    search_results=s.do_search($search_query,$start,$no_of_searches)
    printf "\n[*] Finished searching looking for SQL injections.\n\n"
    s.check_urls(search_results)
rescue Interrupt
   puts "[!] CTRL+C pressed aborting."
   exit(1)
rescue
   puts "[-] Search failed, " + $!.message
   exit(1)
end
